// Copyright Mirela Andronescu 2005-2008
// Takes as input an XML thermo file and creates constraints for CPLEX/IPOPT

// Functions that you need for this file:
//  - A function that returns the number of parameters used in your model
//  - A function that computes how many times each feature occurs in a given structure and the free energy
//  - [optionally] A function that computes the free energy (under your model) of a sequence folded into a given structure.

///////////////////////////////////////////////////////////////////
// MODIFY THIS PART 

// include your own header files
#include "constants.h"
#include "simfold.h"
#include "common.h"
#include "externs.h"
#include "params.h"

// Cristina: May 7
// for HotKnots
#include "init.h"  // for INITIALIZE
#include "initPK.h"  // for INITIALIZE
#include "paramsPK.h"  // for FUNCTION_READ_PARAMETERS_FROM_FILE and FUNCTION_GET_NUMBER_OF_PARAMETERS

// Make sure this number for the maximum length of the sequences is large enough
#define MAXLEN      5000    // maximum length of sequences

// Initialize the data, fill the parameter data structures
// You can write anything here
// "\" separates lines in the define directive
// If you have nothing to initialize, leave it blank:        #define INITIALIZE
#define INITIALIZE \
    char config_file[200] = "../../HotKnotsCommon/HotKnots_v2.0/bin/params/multirnafold.conf";\
    char config_filePK[200] = "../../HotKnotsCommon/HotKnots_v2.0/bin/params/pkenergy.conf";\
    init_data (argv[0], config_file, RNA, 37.0);\
    init_dataPK(argv[0], config_filePK, RNA, 37.0);
    //fill_data_structures_with_new_parameters ("../../../MultiRNAFold/params/turner_parameters_fm363_constrdangles.txt");

// Provide a function that takes as input a file with a set of parameters, and fills your internal data structures
// Example: void fill_data_structures_with_new_parameters (char *parameters_file)
// The parameters_file should have one parameter value per line.
// For example, if you have 3 parameters in your model, with values 1.00, 2.00, and 3.00, the file should be:
// 1.00
// 2.00
// 3.00
// (without the "// ", of course).
#define FUNCTION_READ_PARAMETERS_FROM_FILE  fill_data_structures_with_new_parameters_PK_DP

// Provide the name of a function that returns the number of parameters in your model
// Prototype: int get_num_params()
#define FUNCTION_GET_NUMBER_OF_PARAMETERS  get_num_params_PK_DP


// Specify whether your energy function is LINEAR or QUADRATIC
// *  If your energy function is LINEAR in the energy parameters,
//    then your energy function can be written like this:
//    - deltaG = c' x + f
//        - where x is the vector of parameters
//        - c is the vector of how many times each parameter occurs in the given structure
//        - c' means c transposed
//        - f is a constant
//    - As a simple example, if your model has 3 parameters x1, x2 and x3, and deltaG for some structure is
//      x1 + x3 + x3 + 0.5, then c'=(1,0,2) and f=0.5;
//    - The Turner model underlying mfold, simfold, RNAstructure, RNA Vienna package etc is linear.
// * If your energy function is QUADRATIC in the energy parameters, 
//   then your energy function can be written like this:
//    - deltaG = x' P x + c' x + f
//        - where x is the vector of parameters
//        - P is a symmetric matrix of the coefficients for each quadratic term
//        - c is a vector of counts for each linear term
//        - c' means c transposed
//        - f is a constant
//    - As a simple example, if your model has 3 parameters x1, x2 and x3, and deltaG for some structure is
//      x1*x2 + 0.5*x2*x3 + 2*x2, then P = (0,1,0; 1,0,0.5; 0,0.5,0), c' = (0,2,0) and f=0.
//    - The Dirks&Pierce and Rivas&Eddy models for pseudoknotted structures are quadratic.
//#define ENERGY_FUNCTION_DEGREE  LINEAR
#define ENERGY_FUNCTION_DEGREE  QUADRATIC


// Provide the name of a function that returns the number of occurences (counts) for each feature (see below).
// * If ENERGY_FUNCTION_DEGREE above is LINEAR, the prototype is:
//   double get_feature_counts (char *sequence, char *structure, double *linear_vector, double &free_value);
//      - sequence is a given sequence (input parameter)
//      - structure is a given secondary structure (input parameter)
//      - linear_vector is the c vector above, an array of doubles (output parameter).
//          The function writes the number of occurences for each feature in this vector,
//          for example linear_vector[0] is the number of times the first feature occurs
//      - free_value is the free parameter f above.
//      - assume the memory space was allocated for all input and output variables
//      - return the free_energy of sequence folded into structure. Free_energy should be = c' x + f.
// * If ENERGY_FUNCTION_DEGREE above is QUADRATIC, the prototype is:
//   double get_feature_counts_quadratic (char *sequence, char *structure, double **quadratic_matrix, double *linear_vector, double &free_value);
//      - quadratic_matrix is the P matrix above. Assume the space for it is already allocated.
//        Since the matrix is symmetric, we need only the upper triangle, so only fill it for i <= j.
//      - linear_vector is the c vector above. Assume the space for it is already allocated.
//      - free_value is the free parameter f above.
//      - return the free_energy of sequence folded into structure. Free_energy should be = x' P x + c' x + f.
//#define FUNCTION_GET_FEATURE_COUNTS   get_feature_counts
#define FUNCTION_GET_FEATURE_COUNTS   get_feature_counts_quadratic_PK_DP


// If you have another function which computes the free energy of a sequence folded into a given structure,
//      it is a good idea to check whether FUNCTION_GET_FEATURE_COUNTS returns the same value
// First set HAVE_ALTERNATIVE_FREE_ENERGY_FUNCTION is that's the case
#define HAVE_ALTERNATIVE_FREE_ENERGY_FUNCTION 0

// If the above is 1, then provide the free energy function
//  If the above is 0, leave this unchanged.
// Prototype: double free_energy_simfold (char *sequence, char *structure);
//      returns the free energy (in kcal/mol) of sequence folded into structure
#if (HAVE_ALTERNATIVE_FREE_ENERGY_FUNCTION == 1)
#define FUNCTION_FREE_ENERGY free_energy_PK_DP
#endif 

// That's it, please let me know if it doesn't compile
// DO NOT MODIFY ANYTHING FROM NOW ON!
///////////////////////////////////////////////////////////////////

#define LINEAR  1
#define QUADRATIC 2

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

//#define MAXSLEN 4000
#define ABERANT 99999


int read_params_from_file (char *filename, double *params)
// read parameters (one per row) in filename, and fill the params array.
// return the number of parameters
{
    FILE *file;
    int i = 0;
    char buffer [500];
    if ((file = fopen (filename, "r")) == NULL)
    {
        printf ("Cannot open file %s, ABORT!\n", filename);
        exit(1);
    }
    fgets (buffer, sizeof(buffer), file);
    while (!feof (file))
    {
        sscanf (buffer, "%lf\n", &params[i++]);
        fgets (buffer, sizeof(buffer), file);
    }    
    fclose(file);
    return i;
}


void reset_linear (int num_parameters, double *array)
{
    for (int i=0; i < num_parameters; i++)
    {
        array[i] = 0;
    }
}

void reset_quadratic (int num_parameters, double **array)
{
    for (int i=0; i < num_parameters; i++)
    {
        for (int j=i; j < num_parameters; j++)
        {
            array[i][j] = 0;
        }
    }
}


void my_strcpy (char *source, char *dest, int from, int len)
{
  int i;
  for (i=0; i < len; i++)
    {
      dest[i] = source[from+i];
    }
  dest[i] = '\0';
}

void my_strcat (char *source, char *dest, int from, int len)
{
  int i, l;
  l = strlen (dest);  
  for (i=0; i < len; i++)
    {
      dest[l+i] = source[from+i];
    }
  dest[l+i] = '\0';
}


int main(int argc, char *argv[])
{
    if (argc != 6)
    {
        printf ("Usage: %s <input_xml_filename> <input_parameters_filename> <output_constraints_file> <output_bounds_file> <precision_filename>\n", argv[0]);
        // for the thermo set, the constraints have equalities, so the lower bounds are also upper bounds
        //printf ("Usage: %s <xml_file> <opt_problem_file> --constr-only {1|0} --objective {const|l2norm|l1norm} {MATLAB|CPLEX} <simple_model>\n", argv[0]);
        //printf ("\t num_str    -1 if you want all structures in the file\n");
        exit(0);
    }
    
            
    char filename[200];
    int i, j;
    char sequence[MAXSLEN];
    char structure[MAXSLEN];
    
    // sometimes the initial state is not the completely unfolded sequence. In this case, sequence0 and structure0 are the initial state
    char sequence0[MAXSLEN];
    char structure0[MAXSLEN];    
    
    int self_comp;
    double lb, ub, dg_37, std_dg_37=ABERANT, err_dg_37=ABERANT;
    char buffer [10000];
    // configuration file
    //char config_file[200] = "../HotKnots/simfold/params/multirnafold.conf";
    //char config_file[200] = "../../../MultiRNAFold/params/multirnafold.conf";
    double energy, energy0;
    int param_met[MAXNUMPARAMS];
    char experiment_for_param[MAXNUMPARAMS][100];

    char problem_name[1000];
    char bounds_name[1000];
    char solver[10];
    int res;
    FILE *xml, *problem, *bounds; 
    int constraints_only;
    char objective[20];
    char exp_id[100];
    double precision;
    char precision_filename[1000];
    FILE *precision_file;
    char params_filename[1000];
    
    simple_internal_energy = 0;
    strcpy (filename, argv[1]);
    strcpy (params_filename, argv[2]);
    strcpy (problem_name, argv[3]);
    strcpy (bounds_name, argv[4]);
    strcpy (precision_filename, argv[5]);
    constraints_only = 1;
    strcpy (objective, "l2norm");
    strcpy (solver, "CPLEX");
    simple_internal_energy = 0;
    char constraint_comment[5000];
    char constraint[50000];        

    double *linear_vector;
    double *linear_vector0;
    double free_value;
    double free_value0;
    int pknotted = 0;
    double rmse = 0.0;
    int num = 0;    
    double rmsep = 0.0;     // with pseudoknots
    int nump = 0;
    
    // initialize the thermodynamic parameters
    // call init_data only once for the same dna_or_rna and same temperature
    // if one of them changes, call init_data again (see below)
    INITIALIZE;  

    FUNCTION_READ_PARAMETERS_FROM_FILE (params_filename);
       
    //num_params = create_string_params();   
    // HOTKNOTS
    // I need to know how many parameters there are in the model
    int num_parameters = FUNCTION_GET_NUMBER_OF_PARAMETERS();

    double *params_array;
    params_array = new double[num_parameters];
    int n = read_params_from_file (params_filename, params_array);
    if (n != num_parameters)
    {
        printf ("ERROR! Number of parameters in file %s is %d and is NOT the same as what the function %s reported: %d!\n",
            params_filename, n, FUNCTION_GET_NUMBER_OF_PARAMETERS, num_parameters);
        exit(1);
    }
    // now params_array is filled with the values in the given parameter file

    // Allocate the space for the counters
    linear_vector = new double[num_parameters];
    if (linear_vector == NULL)
    {
        printf ("ERROR! Space could not be allocated for linear_vector, ABORT!\n");
        exit(1);
    }
    linear_vector0 = new double[num_parameters];
    if (linear_vector0 == NULL)
    {
        printf ("ERROR! Space could not be allocated for linear_vector0, ABORT!\n");
        exit(1);
    }

    // allocate the two dimensional array if applicable
    #if (ENERGY_FUNCTION_DEGREE == QUADRATIC)

    double **quadratic_matrix;
    quadratic_matrix = new double*[num_parameters];
    if (quadratic_matrix == NULL)
    {
        printf ("ERROR! Space could not be allocated for quadratic_matrix, ABORT!\n");
        exit(1);
    }
    for (i=0; i < num_parameters; i++)
    {
        // I tried to allocate only i+1 elements, but it crashes during delete - dunno why
        quadratic_matrix[i] = new double[num_parameters];
        if (quadratic_matrix[i] == NULL)
        {
            printf ("ERROR! Space could not be allocated for quadratic_matrix[%d], ABORT!\n", i);
            exit(1);
        }
    }
    double **quadratic_matrix0;
    quadratic_matrix0 = new double*[num_parameters];
    if (quadratic_matrix0 == NULL)
    {
        printf ("ERROR! Space could not be allocated for quadratic_matrix0, ABORT!\n");
        exit(1);
    }
    for (i=0; i < num_parameters; i++)
    {
        // I tried to allocate only i+1 elements, but it crashes during delete - dunno why
        quadratic_matrix0[i] = new double[num_parameters];
        if (quadratic_matrix0[i] == NULL)
        {
            printf ("ERROR! Space could not be allocated for quadratic_matrix0[%d], ABORT!\n", i);
            exit(1);
        }
    }    
    #endif

    if ((problem = fopen (problem_name,"w")) == NULL)
    {
      printf ("Cannot open file %s\n", problem_name);
      return 0;
    }
    
    if ((bounds = fopen (bounds_name,"w")) == NULL)
    {
      printf ("Cannot open file %s\n", bounds_name);
      return 0;
    }
    
    if ((xml = fopen (filename, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", filename);
        exit (0);
    }           
    
    int k;
    k = 0;
    int delta_index = 1;
    double margin;
    int num_experiments;    // = 686;    //    506;    //524;
    
    num_experiments = 0;
    // get num_experiments
    while (!feof (xml))
    {    
        // the file is in XML format, among whose lines it has the following types:
        //    <EXPERIMENT sequence="CCGG CCGG" structure="(((( ))))" dG-37="-4.36" STD-dG-37="0.1"> </EXPERIMENT>
        //    <EXPERIMENT sequence="UGACCUCA UGAGGUCA" structure="(((((((( ))))))))" dG-37="-12.34" ERR-dG-37="3%"> </EXPERIMENT>
        char *begin;
        fgets (buffer, sizeof(buffer), xml);
        begin = strstr (buffer, "<EXPERIMENT");        
        if (begin == NULL) continue;    
        num_experiments++;
    }
    fclose (xml);
    
    //printf ("Num exp: %d\n", num_experiments);
    
    if ((xml = fopen (filename, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", filename);
        exit (0);
    }               
    
    if ((precision_file = fopen (precision_filename,"w")) == NULL)
    {
      printf ("Cannot open file %s\n", precision_filename);
      return 0;
    }
    
    while (!feof (xml))
    {    
        // the file is in XML format, among whose lines it has the following types:
        //    <EXPERIMENT sequence="CCGG CCGG" structure="(((( ))))" dG-37="-4.36" STD-dG-37="0.1"> </EXPERIMENT>
        //    <EXPERIMENT sequence="UGACCUCA UGAGGUCA" structure="(((((((( ))))))))" dG-37="-12.34" ERR-dG-37="3%"> </EXPERIMENT>
        
        // try including only the first paper. With everything I can't save the solution because it says "No basic solution"
        //    Problem solved: In the cplex script, I have to write: "write file.vec"
        //if (counter >= 94) break;
        
        fgets (buffer, sizeof(buffer), xml);
        //printf ("%s", buffer);                         
        char *begin;
        char temp[100];
        sequence[0] = '\0'; structure[0] = '\0'; dg_37 = ABERANT; err_dg_37 = ABERANT; std_dg_37 = ABERANT;
        begin = strstr (buffer, "<EXPERIMENT");        
        if (begin == NULL) continue;
        
        sequence0[0] = '\0';
        structure0[0] = '\0';
        get_data_from_buffer (buffer, "id=\"", '\"', exp_id);
        get_data_from_buffer (buffer, "sequence0=\"", '\"', sequence0);        
        get_data_from_buffer (buffer, "structure0=\"", '\"', structure0);        
        get_data_from_buffer (buffer, "sequence=\"", '\"', sequence);
        get_data_from_buffer (buffer, "structure=\"", '\"', structure);
        get_data_from_buffer (buffer, "dG-37=\"", '\"', temp);  dg_37 = atof (temp);
        if (get_data_from_buffer (buffer, "STD-dG-37=\"", '\"', temp))  std_dg_37 = atof (temp);
        if (get_data_from_buffer (buffer, "ERR-dG-37=\"", '%', temp))  err_dg_37 = atof (temp);                                      
        
        //printf ("=======\n%s\n%s\t%s\n", exp_id, sequence, structure);
        
        // Self-complementarity doesn't have to do anything with dG. Symmetry does! 
        // If symmetric, the 0.43 penalty is part of the free energy value.
        // Not exactly sure what symmetric means.
        //  Following conversations with Dave Mathews,
        //      I consider that 2 sequences are symmetric if they are perfectly complementary and identical
        
        
        // first, adjust dg_37 according to self-complementarity
        //  Let S1 be the first sequence, let S2 be the second, and let S3 be the reverse complement of S1. 
        //    I considered the S1/S2 duplex is self-complementary if S1, S2 and S3 are identical.
        // This is incorect!! The 0.43 penalty has to be added when S1 and S2 are identical,
        //      even if they are not perfectly complementary.
        char *space;
        int self_comp = 0;
        //char rev_comp_of_first_part[MAXSLEN/2];        
        space = strstr (sequence, " ");
        if (space != NULL)
        {
            int space_position;
            space_position = space - sequence;
            char first_part[MAXSLEN/2];
            strcpy (first_part, sequence);
            first_part[space_position] = '\0';
            //reverse_complement_of_seq (first_part, rev_comp_of_first_part);
            // Mirela comment added on Dec 14, 2007: the self-complementary sequences can have mismatches, so the following was incorrect
            // if (strcmp (first_part, space+1) == 0 && strstr (structure, ".") == NULL)
            if (strcmp (first_part, space+1) == 0)  // && strstr (structure, ".") == NULL)
            // they are identical and are not perfectly complementary
            //if (strcmp (first_part, space+1) == 0 && strcmp (first_part, rev_comp_of_first_part) == 0)
            {
                //printf ("self-comp: %s\t%s\n", sequence, structure);
                dg_37 -= 0.43;
                self_comp = 1;
            }    
        }
        
        // compute the lower bound and upper bound
        if (err_dg_37 < ABERANT && std_dg_37 < ABERANT)
        {
            printf ("One of err_dg_37 and std_dg_37 hs to be ABERANT. err_dg_37 = %.2lf, std_dg_37 = %.2lf\n", err_dg_37, std_dg_37);
            exit (1);
        }
        if (err_dg_37 < ABERANT)
        {
            //margin = dg_37 * err_dg_37/100;    // it gets infeasible
            margin = dg_37 * err_dg_37/100;
            if (margin < 0) margin *= -1;
            std_dg_37 = fabs(dg_37) * err_dg_37/100;
        }
        else
        {
            // I assume a normal distribution
            // Thus 99.73% of the area is within three standard deviations.
            //margin = std_dg_37*3; // this gets unfeasible, try with 4
            margin = std_dg_37*3;
        }
        lb = dg_37 - margin;
        ub = dg_37 + margin;
      
        // Added on Mar 20, 2008: the minimum error is 0.1
        if (std_dg_37 < 0.1)   std_dg_37 = 0.1;
        precision = 1/(2*std_dg_37*std_dg_37);
        //precision = 1;
        
        //printf ("%s std = %g\n", exp_id, std_dg_37);
        fprintf (precision_file, "%.15lf\n", precision);


        if (strcmp (sequence0, "") != 0)
        {                        
            reset_linear (num_parameters, linear_vector0);
            #if (ENERGY_FUNCTION_DEGREE == LINEAR)
            energy0 = FUNCTION_GET_FEATURE_COUNTS (sequence0, structure0, linear_vector0, free_value0);
            if (!check_counts_linear (num_parameters, params_array, linear_vector0, free_value0, energy0))
            {
                // message is printed from check_counts_linear
                printf ("%s\n%s\n", sequence0, structure0);
                exit(1);
            }
            #elif (ENERGY_FUNCTION_DEGREE == QUADRATIC)
            reset_quadratic (num_parameters, quadratic_matrix0);
            // HotKnots can't deal with interacting molecules
            if (strstr(structure0,"[") == NULL)     // non-pseudoknotted
            {
                energy0 = count_each_structure_type (sequence0, structure0, "", linear_vector0, free_value0, 1);
            }
            else
            {
                energy0 = FUNCTION_GET_FEATURE_COUNTS (sequence0, structure0, quadratic_matrix0, linear_vector0, free_value0);
                if (!check_counts_quadratic (num_parameters, params_array, quadratic_matrix0, linear_vector0, free_value0, energy0))
                {
                    // message is printed from check_counts_linear
                    printf ("%s\n%s\n", sequence0, structure0);
                    //exit(1);
                }
            }
            #else
            printf ("ERROR! Constant ENERGY_FUNCTION_DEGREE should be either LINEAR or QUADRATIC! Abort.\n");
            exit(1);
            #endif
    
            // also check if energy_fe is the same as the one provided separately
            #if (HAVE_ALTERNATIVE_FREE_ENERGY_FUNCTION == 1)
            alt_energy = FUNCTION_FREE_ENERGY (sequence0, structure0);
            if (fabs (alt_energy - energy0) > 10.0)
            {
                printf ("dG by %s is different from dG by %s: %.2lf vs. %.2lf for \n%s\n%s\n",
                    FUNCTION_FREE_ENERGY, FUNCTION_GET_FEATURE_COUNTS, alt_energy, energy0, sequence0, structure0);
                exit(1);
            }                                       
            #endif         
        

        }
        // added on Apr 17, 2008
        else
        {
            reset_linear (num_parameters, linear_vector0);
            #if (ENERGY_FUNCTION_DEGREE == QUADRATIC)
            reset_quadratic (num_parameters, quadratic_matrix0);
            #endif
            free_value0 = 0;
            energy0 = 0.0;
        }
        
        // now I have the real sequence and structure
        
        //count_each_structure_type (sequence, structure, "", counter_min, f_min, 1);
        reset_linear (num_parameters, linear_vector);
        #if (ENERGY_FUNCTION_DEGREE == LINEAR)
        energy = FUNCTION_GET_FEATURE_COUNTS (sequence, structure, linear_vector, free_value);
        if (!check_counts_linear (num_parameters, params_array, linear_vector, free_value, energy))
        {
            // message is printed from check_counts_linear
            printf ("%s\n%s\n", sequence, structure);
            exit(1);
        }
        #elif (ENERGY_FUNCTION_DEGREE == QUADRATIC)
        reset_quadratic (num_parameters, quadratic_matrix);
        // HotKnots can't deal with interacting molecules
        if (strstr(structure,"[") == NULL)     // non-pseudoknotted
        {
            energy = count_each_structure_type (sequence, structure, "", linear_vector, free_value, 1);
            pknotted = 0;
            num++;
            rmse += (dg_37-energy+energy0)*(dg_37-energy+energy0);
        }
        else
        {
            energy = FUNCTION_GET_FEATURE_COUNTS (sequence, structure, quadratic_matrix, linear_vector, free_value);
            if (!check_counts_quadratic (num_parameters, params_array, quadratic_matrix, linear_vector, free_value, energy))
            {
                // message is printed from check_counts_linear
                printf ("%s\n%s\n", sequence, structure);
                //exit(1);
            }
            pknotted = 1;
            nump++;
            rmsep += (dg_37-energy+energy0)*(dg_37-energy+energy0);
        }
        #else
        printf ("ERROR! Constant ENERGY_FUNCTION_DEGREE should be either LINEAR or QUADRATIC! Abort.\n");
        exit(1);
        #endif

        // also check if energy_fe is the same as the one provided separately
        #if (HAVE_ALTERNATIVE_FREE_ENERGY_FUNCTION == 1)
        alt_energy = FUNCTION_FREE_ENERGY (sequence, structure);
        if (fabs (alt_energy - energy) > 10.0)
        {
            printf ("dG by %s is different from dG by %s: %.2lf vs. %.2lf for \n%s\n%s\n",
                FUNCTION_FREE_ENERGY, FUNCTION_GET_FEATURE_COUNTS, alt_energy, energy, sequence, structure);
            exit(1);
        }
        #endif   


        

        // do correction for Giese_Serra_1998 structures: the dangling ends should actually be tstackh, otherwise it's not full rank
/*        
        if ((strstr (exp_id, "Giese_Serra_1998") != NULL) ||
            (strstr (exp_id, "Dale_Serra_2000") != NULL) ||
            (strstr (exp_id, "Serra_Turner_1994") != NULL) ||
            (strstr (exp_id, "Vecenie_Serra_2004") != NULL) ||
            (strstr (exp_id, "Vecenie_Serra_2006") != NULL) )
        {         
            int bi,bj,bk,bl;   
            char *b;
            char stype[50];
            int index;
            int to_be_adjusted = 0;
            for (i=0; i < num_params; i++)
            {
                if ( (b = strstr (string_params[i], "dangle")) != NULL && counter_min[i] > 0)
                {
                    to_be_adjusted = 1;
                    counter_min[i] = 0;
                    // has dangle
                    //printf ("*** %s ", string_params[i]);
                    if (b[7] == 't')    // it's top
                    {
                        if (b[11] == '0')         bi=0;
                        else if (b[11] == '1')    bi=1;
                        else if (b[11] == '2')    bi=2;
                        else if (b[11] == '3')    bi=3;
                        if (b[14] == '0')         bj=0;
                        else if (b[14] == '1')    bj=1;
                        else if (b[14] == '2')    bj=2;
                        else if (b[14] == '3')    bj=3;
                        if (b[17] == '0')         bk=0;
                        else if (b[17] == '1')    bk=1;
                        else if (b[17] == '2')    bk=2;
                        else if (b[17] == '3')    bk=3;
                                                
                        //printf ("top i=%d j=%d k=%d ", bi, bj, bk);
                    }
                    else
                    {
                        if (b[17] == '0')         bl=0;
                        else if (b[17] == '1')    bl=1;
                        else if (b[17] == '2')    bl=2;
                        else if (b[17] == '3')    bl=3;                        
                        //printf ("bot ");
                    }
                    //break;                    
                } 
            }
            if (to_be_adjusted)
            {
                sprintf (stype, "tstackh[%d][%d][%d][%d]", bi, bj, bk, bl);
                //printf (" %s ", stype);
                index = structure_type_index (stype);
                counter_min [index] = 2;
                //printf (" %d %s %s\n", index, sequence, structure);
            }                
        }
*/        

        // write for IPOPT
        sprintf (constraint, " ");
        int l=0;

        #if (ENERGY_FUNCTION_DEGREE == QUADRATIC)
        for (i=0; i < num_parameters; i++)
        {
            for (j=i; j < num_parameters; j++)
            {
                if (quadratic_matrix[i][j] - quadratic_matrix0[i][j] != 0)
                {
                    if (l > 0)  sprintf (constraint, "%s +", constraint);
                    if (quadratic_matrix[i][j]-quadratic_matrix0[i][j] > 0)
                    {
                        sprintf (constraint, "%s %g * x[%d] * x[%d] ", constraint, quadratic_matrix[i][j]-quadratic_matrix0[i][j], i, j);
                    }
                    else
                    {
                        sprintf (constraint, "%s (%g) * x[%d] * x[%d] ", constraint, quadratic_matrix[i][j]-quadratic_matrix0[i][j], i, j);
                    }
                    l++;
                }
            }
        }
        #endif

        for (i=0; i < num_parameters; i++)
        {
            if (linear_vector[i] - linear_vector0[i] != 0)
            {
                if (l > 0)  sprintf (constraint, "%s +", constraint);
                if (linear_vector[i]-linear_vector0[i] > 0)
                {
                    sprintf (constraint, "%s %.1lf * x[%d]", constraint, linear_vector[i]-linear_vector0[i], i);
                }
                else
                {
                    sprintf (constraint, "%s (%.1lf) * x[%d]", constraint, linear_vector[i]-linear_vector0[i], i);
                }
                l++;
            }
        }

        fprintf (problem, "%s\n", constraint);
        
        
        // compute RMSE
        printf ("%s, exp en: %.2lf, estimated: %.2lf, pknotted: %d\n", exp_id, dg_37, energy-energy0, pknotted);
        
        
        

        // check if there is any free parameter
        double diff_free;
        diff_free = free_value - free_value0;
        //sprintf (constraint_comment, "%s + %.2lf = %.2lf ", constraint_comment, diff_free/100.0, dg_37);
        if (diff_free != 0)
        {
            dg_37 -= diff_free/100.0;
        }
        fprintf (bounds, "%.2lf\n", dg_37);


        /*
        char lb_constraint[5000];  
              
        if (strcmp (solver, "CPLEX") ==0)       // ACTUALLY IT'S FOR IPOPT
        {
            int l=0;
            // first write the lower bound
            //sprintf (lb_constraint, "    tc_%s: ", exp_id);    // thermodynamic constraints
            // get the id only, it's too long otherwise
            char idid[10];
            for (int ii=0; ii < strlen (exp_id); ii++)
            {
                if (exp_id[ii] == '_')
                {
                    idid[ii] = '\0';
                    break;
                }
                idid[ii] = exp_id[ii];
            }
            //sprintf (lb_constraint, "    tc_%s: ", idid);    // thermodynamic constraints
            sprintf (lb_constraint, " ");    // thermodynamic constraints
            sprintf (constraint_comment,      "    \\* ");
            int newlines_in_comment = 1;
            for (i=0; i < num_params; i++)
            {
                if (strlen(constraint_comment) > newlines_in_comment*400)
                {
                    sprintf (constraint_comment, "%s*\\\n    \\*", constraint_comment);
                    newlines_in_comment++;
                }
                
                if (counter_min[i] - counter0[i] != 0)
                {
                    param_met[i] = 1;
                    if (experiment_for_param[i][0] == '\0')
                        strcpy (experiment_for_param[i], exp_id);
                    if (l > 0)
                    { 
                        sprintf (lb_constraint, "%s +", lb_constraint);
                        sprintf (constraint_comment, "%s + ", constraint_comment);
                    }   
                    // don't write the counter if it's 1    
                    if (counter_min[i]-counter0[i] > 0)
                    {
                        sprintf (lb_constraint, "%s %.1lf * x[%d]", lb_constraint,counter_min[i]-counter0[i], i);
                        sprintf (constraint_comment, "%s %s ", constraint_comment, string_params[i]);
                    }    
                    else
                    {
                        sprintf (lb_constraint, "%s (%.1lf) * x[%d] ", lb_constraint, counter_min[i]-counter0[i], i);
                        sprintf (constraint_comment, "%s %.1lf %s ", constraint_comment, counter_min[i]-counter0[i], string_params[i]);
                    }    
                    //if (counter_min[i] == 1)  fprintf (problem, " x%d (%s) ", i, string_params[i]);   
                    //else fprintf (problem, " %d x%d (%s) ", counter_min[i], i, string_params[i]);
                    l++;
//                     if (l % 10 == 0)
//                     {
//                         sprintf (lb_constraint, "%s\n        ", lb_constraint);
//                     }

                }
            }    
            
            // check if there is any free parameter
            double diff_free;
            diff_free = free_value - free_value0;
            //sprintf (constraint_comment, "%s + %.2lf = %.2lf ", constraint_comment, diff_free/100.0, dg_37);
            if (diff_free != 0)
            {
                dg_37 -= diff_free/100.0;
            }                        
            
            //sprintf (constraint_comment, "%s *\\\n", constraint_comment);
//             if (strcmp (objective, "const") == 0)    
//             {
//                 sprintf (lb_constraint, "%s >= %.2lf\n", lb_constraint, lb);
//             }
//             else     // if l1norm or l2norm    
//             {
//                 sprintf (lb_constraint, "%s - v%d = %.2lf\n", lb_constraint, counter, dg_37);
//             }

            fprintf (bounds, "%.2lf\n", dg_37);
            counter++;
            //fprintf (problem, "%s\n", constraint_comment);
            fprintf (problem, "%s\n", lb_constraint);
                       
        }    // end if CPLEX
        */
        
        //fclose (problem);
        //exit(1);
    }
    
    rmse /= num;
    rmsep /= nump;
    rmse = sqrt(rmse);
    rmsep = sqrt(rmsep);
    printf ("num = %d, rmse = %.3lf\n", num, rmse);
    printf ("nump= %d, rmsep= %.3lf\n", nump, rmsep);
    
    fclose (problem);
    fclose (bounds);
    fclose (xml);
    fclose (precision_file);
    return 0;
}
