// Copyright Mirela Andronescu 2005-2008
// Predicts and outputs minimum free energy secondary structures with new parameters;
//  also outputs the accuracy of the new predictions
// Takes as input:
//  - a set of new parameters
//  - a set of sequences with known and initial structures
//  Outputs:
//  - a file with sequences, known and previously predicted structures, and the newly predicted structures
//  - a file with the sensitivity and positive predictive values for the newly predicted structures    

// You need the following functions (see below for details):
//  - a function that takes as input a file with a set of parameters, and fills your internal data structures
//  - a function that predicts the minimum free energy (or low free energy) secondary structure for a given sequence
//  - if your data has secondary structure constraints, you need a variant of the
//          prediction function to accomodate constraints
//  - a function that computes the sensitivity of prediction
//  - a function that computes the positive predictive value of the prediction

///////////////////////////////////////////////////////////////////
// MODIFY THIS PART 

// Include your own header files
#include "simfold.h"
#include "common.h"
#include "params.h"
#include "externs.h"

// Cristina: May 7
// for HotKnots
#include "init.h"  // for INITIALIZE
#include "initPK.h"  // for INITIALIZE
#include "paramsPK.h"  // for FUNCTION_READ_PARAMETERS_FROM_FILE
#include "commonPK.h"  // for sensitivity and ppv
#include "HotKnotEnergy.h"  // for FUNCTION_PREDICTOR

// Make sure this number for the maximum length of the sequences is large enough
#define MAXLEN      5000    // maximum length of sequences

// Cristina: Aug. 18 - added this 
#define MAXSUBOPTS      21    // maximum number of suboptimal + optimal structures

// Initialize the data, fill the parameter data structures
// You can write anything here
// "\" separates lines in the define directive
#define INITIALIZE \
    char config_file[200] = "../../HotKnotsCommon/HotKnots_v2.0/bin/params/multirnafold.conf";\
    char config_filePK[200] = "../../HotKnotsCommon/HotKnots_v2.0/bin/params/pkenergy.conf";\
    init_data (argv[0], config_file, RNA, 37.0);\
    init_dataPK(argv[0], config_filePK, RNA, 37.0);

// Provide a function that takes as input a file with a set of parameters, and fills your internal data structures
// Example: void fill_data_structures_with_new_parameters (char *parameters_file)
// The parameters_file should have one parameter value per line.
// For example, if you have 3 parameters in your model, with values 1.00, 2.00, and 3.00, the file should be:
// 1.00
// 2.00
// 3.00
// (without the "// ", of course).
#define FUNCTION_READ_PARAMETERS_FROM_FILE  fill_data_structures_with_new_parameters_PK_DP

// Cristina: Aug. 18
// Set to 1 if you want to print suboptimals
#define PRINT_SUBOPTIMALS 1

// If there is a chance there might be structures in the training or testing data sets
//  that DO NOT have structural constraints, then set UNCONSTRAINTED_STRUCTURES to 1
#define UNCONSTRAINED_STRUCTURES  1

// If the above is 1, then you have to provide a function that predicts
//  the minimum free energy (or a low free energy) secondary structure. Otherwise ignore this.
#if (UNCONSTRAINED_STRUCTURES == 1)
    // Prototype: double simfold (char *sequence, char *structure)
    //      - sequence is the given RNA sequence to be folded (input parameter)
    //      - structure is the predicted secondary structure (output parameter)
    //      - the function may return the energy, or return nothing, it doesn't matter
	#if (PRINT_SUBOPTIMALS == 0)
	    #define FUNCTION_PREDICTOR hotknots_DP
	#endif
    // Prototype: double simfold (char *sequence, char **structure, double* energies)
	//		- energies is an array of the energies of the suboptimals
	#if (PRINT_SUBOPTIMALS == 1)
	    #define FUNCTION_PREDICTOR hotknots_DP_suboptimals
	#endif
#endif

// If there is a chance there might be structures in the training or testing data sets
//  that DO have structural constraints, then set CONSTRAINTED_STRUCTURES to 1
// Note that the training or testing sets might have both constrained and unconstrained structures
#define CONSTRAINED_STRUCTURES  0

// If CONSTRAINED_STRUCTURES is 1, then you have to provide a variant of the prediction function
//  that takes the restricted string as input
// If CONSTRAINED_STRUCTURES is 0, it doesn't matter what you write here
#if (CONSTRAINED_STRUCTURES == 1)
    // Prototype: double simfold_restricted (char *sequence, char *restricted, char *structure)
    //      - restricted is a given restriction string
    //          (dots and parentheses for the restricted parts, _ (underscore) for unrestricted)
	#if (PRINT_SUBOPTIMALS == 0)
	    #define FUNCTION_PREDICTOR_CONSTRAINED simfold_restricted
	#endif
	#if (PRINT_SUBOPTIMALS == 1)
	    #define FUNCTION_PREDICTOR_CONSTRAINED simfold_restricted
	#endif
#endif

// Provide a function that computes the sensitivity of the predicted structure vs. the reference structure
// The definition that I use is: number of correctly predicted base pairs / number of true base pairs
//  Note: if number of true base pairs is 0, I return -1.
// Prototype: double compute_sensitivity (char *reference_structure, char *predicted_structure)
//      - reference_structure is the reference structure, considered ground truth
//      - predicted_structure is some predicted structure
//      - returns sensitivity in [0,1] or -1 if denominator is 0
#define FUNCTION_COMPUTE_SENSITIVITY compute_PK_sensitivity

// Provide a function that computes the positive predictive value (ppv) of the predicted structure
//  vs. the reference structure
// The definition that I use is: number of correctly predicted base pairs / number of predicted base pairs
//  Note: if number of predicted base pairs is 0, I return -1.
// Prototype: double compute_ppv (char *reference_structure, char *predicted_structure)
//      - reference_structure is the reference structure, considered ground truth
//      - predicted_structure is some predicted structure
//      - returns ppv in [0,1] or -1 if denominator is 0
#define FUNCTION_COMPUTE_PPV compute_PK_ppv

// PLEASE LET ME KNOW IF YOU WANT TO USE OTHER ACCURACY FUNCTIONS




// I need the following 5 defines to compute the energy of the known structure
// For Hotknots, the only correct function to do that was the function that also computes the counts
// So that's what I'm going to use. Otherwise, you can just use a direct function

// Provide the name of a function that returns the number of parameters in your model
// Prototype: int get_num_params()
#define FUNCTION_GET_NUMBER_OF_PARAMETERS  get_num_params_PK_DP

// Specify whether your energy function is LINEAR or QUADRATIC
// *  If your energy function is LINEAR in the energy parameters,
//    then your energy function can be written like this:
//    - deltaG = c' x + f
//        - where x is the vector of parameters
//        - c is the vector of how many times each parameter occurs in the given structure
//        - c' means c transposed
//        - f is a constant
//    - As a simple example, if your model has 3 parameters x1, x2 and x3, and deltaG for some structure is
//      x1 + x3 + x3 + 0.5, then c'=(1,0,2) and f=0.5;
//    - The Turner model underlying mfold, simfold, RNAstructure, RNA Vienna package etc is linear.
// * If your energy function is QUADRATIC in the energy parameters, 
//   then your energy function can be written like this:
//    - deltaG = x' P x + c' x + f
//        - where x is the vector of parameters
//        - P is a symmetric matrix of the coefficients for each quadratic term
//        - c is a vector of counts for each linear term
//        - c' means c transposed
//        - f is a constant
//    - As a simple example, if your model has 3 parameters x1, x2 and x3, and deltaG for some structure is
//      x1*x2 + 0.5*x2*x3 + 2*x2, then P = (0,1,0; 1,0,0.5; 0,0.5,0), c' = (0,2,0) and f=0.
//    - The Dirks&Pierce and Rivas&Eddy models for pseudoknotted structures are quadratic.
//#define ENERGY_FUNCTION_DEGREE  LINEAR
#define ENERGY_FUNCTION_DEGREE  QUADRATIC


// Provide the name of a function that returns the number of occurences (counts) for each feature (see below).
// * If ENERGY_FUNCTION_DEGREE above is LINEAR, the prototype is:
//   double get_feature_counts (char *sequence, char *structure, double *linear_vector, double &free_value);
//      - sequence is a given sequence (input parameter)
//      - structure is a given secondary structure (input parameter)
//      - linear_vector is the c vector above, an array of doubles (output parameter).
//          The function writes the number of occurences for each feature in this vector,
//          for example linear_vector[0] is the number of times the first feature occurs
//      - free_value is the free parameter f above.
//      - assume the memory space was allocated for all input and output variables
//      - return the free_energy of sequence folded into structure. Free_energy should be = c' x + f.
// * If ENERGY_FUNCTION_DEGREE above is QUADRATIC, the prototype is:
//   double get_feature_counts_quadratic (char *sequence, char *structure, double **quadratic_matrix, double *linear_vector, double &free_value);
//      - quadratic_matrix is the P matrix above. Assume the space for it is already allocated.
//        Since the matrix is symmetric, we need only the upper triangle, so only fill it for i <= j.
//      - linear_vector is the c vector above. Assume the space for it is already allocated.
//      - free_value is the free parameter f above.
//      - return the free_energy of sequence folded into structure. Free_energy should be = x' P x + c' x + f.
#define FUNCTION_GET_FEATURE_COUNTS   get_feature_counts_quadratic_PK_DP



// That's it, please let me know if it doesn't compile
// DO NOT MODIFY ANYTHING FROM NOW ON!
///////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#define LINEAR  1
#define QUADRATIC 2

void create_filename (char *input_filename, char *output_filename, int &seqid)
// input_filename is path/name.txt
// output_filename is path/tmp/name_seqid
// ASSUME path/tmp exists
{
    int i;
    int position = 0;
    char justname[100];
    for (i=0; i < strlen(input_filename); i++)
        if (input_filename[i] == '/')
            position = i;
    // now copy in output_filename path
    position++;
    strncpy (output_filename, input_filename, position);
    // add "tmp/"
    output_filename[position++] = 't';
    output_filename[position++] = 'm';
    output_filename[position++] = 'p';
    output_filename[position++] = '/';
    for (i=position-4; i < strlen(input_filename); i++)
    {
        // stop at the dot
        if (input_filename[i] == '.')
        {
            output_filename[position] = '\0';
            break;
        }
        output_filename[position++] = input_filename[i];
    }
    sprintf (output_filename, "%s_%d_", output_filename, seqid);
    seqid++;
}


int identical (char* structure1, char *structure2)
// check if the following two structures are actually identical
// for example the following are identical, but their strings are different:
// AAAAAACUAAUAGAGGGGGGACUUAGCGCCCCCCAAACCGUAACCCC
// ..............((((((.....[[[))))))....]]]......
// ..............[[[[[[.....(((]]]]]]....)))......
// return 1 if they are identical, 0 otherwise
{
    int len = strlen (structure1);
    short ptable1[len];
    short ptable2[len];
    detect_original_PKed_pairs_many (structure1, ptable1);
    detect_original_PKed_pairs_many (structure2, ptable2);
    for (int i=0; i < len; i++)
    {
        if (ptable1[i] != ptable2[i])   return 0;
    }
    return 1;
}

double get_energy (char *sequence, char *structure)
// given a sequence and structure as input, return the free energy with the given parameters
// This function should just call a function that exists in the prediction program
// For Hotknots, that function is not correct at the moment, so I'm using the function
//      that computes the counts and also returns the free energy   
{
    double *linear_vector;
    double free_value;
    int i;
    double energy;
    
    // I need to know how many parameters there are in the model
    int num_parameters = FUNCTION_GET_NUMBER_OF_PARAMETERS();    
    
    // Allocate the space for the counters
    linear_vector = new double[num_parameters];
    if (linear_vector == NULL)
    {
        printf ("ERROR! Space could not be allocated for linear_vector, ABORT!\n");
        exit(1);
    }    
    
    // allocate the two dimensional array if applicable
    #if (ENERGY_FUNCTION_DEGREE == QUADRATIC)
    double **quadratic_matrix;
    quadratic_matrix = new double*[num_parameters];
    if (quadratic_matrix == NULL)
    {
        printf ("ERROR! Space could not be allocated for quadratic_matrix, ABORT!\n");
        exit(1);
    }
    for (i=0; i < num_parameters; i++)
    {
        // I tried to allocate only i+1 elements, but it crashes during delete - dunno why
        quadratic_matrix[i] = new double[num_parameters];
        if (quadratic_matrix[i] == NULL)
        {
            printf ("ERROR! Space could not be allocated for quadratic_matrix[%d], ABORT!\n", i);
            exit(1);
        }
    }    
    #endif
    
    #if (ENERGY_FUNCTION_DEGREE == LINEAR)
    energy = FUNCTION_GET_FEATURE_COUNTS (sequence, structure, linear_vector, free_value);
    #elif (ENERGY_FUNCTION_DEGREE == QUADRATIC)
    energy = FUNCTION_GET_FEATURE_COUNTS (sequence, structure, quadratic_matrix, linear_vector, free_value);
    #else
    printf ("ERROR! Constant ENERGY_FUNCTION_DEGREE should be either LINEAR or QUADRATIC! Abort.\n");
    exit(1);
    #endif
        
    delete [] linear_vector;
    #if (ENERGY_FUNCTION_DEGREE == QUADRATIC)
    for (i=0; i < num_parameters; i++)  delete [] quadratic_matrix[i];
    delete [] quadratic_matrix;    
    #endif
    return energy;
}        



int main (int argc, char *argv[])
{
    if (argc != 4)
    {
        printf ("Usage: %s <input_new_params_file> <input_old_set_file> <output_new_set_file>\n", argv[0]);
        // write predicted structure always
        // write the accuracy in the same file
        exit(0);
    }    

    char sequence[MAXLEN];
    char known_structure[MAXLEN];
    char restricted[MAXLEN];
    char initial_structure[MAXLEN];
    char new_structure[MAXLEN];  // Cristina: no longer used
    char** new_structures;  // Cristina: Aug. 18    
    double energies[MAXSUBOPTS];  // Cristina: Aug. 18 - energies for suboptimal structure predictions
    double energy;

	new_structures = new char*[MAXSUBOPTS];
	// initialize all structures to null strings, so can terminate loop later on
	for (int i = 0; i < MAXSUBOPTS; i++)
	{
		new_structures[i] = new char[MAXLEN];
//		new_structures[i] = "\0";
	}

    char new_params_filename[200];
    char oldset_filename[200];
    char newset_filename[200];
    FILE *oldset_file, *newset_file;
    int i, j, k;
    char buffer[MAXLEN];

    // check if simfold prediction is better
    char simfold_structure[MAXLEN];
    double energy_simfold;
    
    char accuracy_output_filename[500];
    FILE *accuracy_output_file;
    
    strcpy (new_params_filename, argv[1]);
    strcpy (oldset_filename, argv[2]);
    strcpy (newset_filename, argv[3]);

    INITIALIZE;

    FUNCTION_READ_PARAMETERS_FROM_FILE (new_params_filename);

    j = 0;
    double sens_initial_train, sens_new_train;
    double ppv_initial_train, ppv_new_train;    

  //////////////////////////////////////////////////////
  // open the old set, to get the sequence, the true structure and the turner structure
    if ((oldset_file = fopen(oldset_filename, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", oldset_filename);
        exit(0);
    }
    
    if ((newset_file = fopen(newset_filename, "w")) == NULL)
    {
        printf ("Cannot open file %s\n", newset_filename);
        exit(0);
    }
   
    //the format is:
    //    > header
    //    sequence
    //    known structure
    //    [restricted]
    //    initial structure
    //    pred structure 1
    //    pred structure 2 etc
    //    empty line    
    
    // read the header line here
    fgets (buffer, sizeof(buffer), oldset_file);    // header "> SSTRAND_ID etc"
    int seqid = 0;
    while (!feof(oldset_file))
    {              
        fprintf (newset_file, "%s", buffer);
        fgets (sequence, sizeof(buffer), oldset_file);    // sequence
        fprintf (newset_file, "%s", sequence);
        fgets (known_structure, sizeof(buffer), oldset_file);    // known structure
        fprintf (newset_file, "%s", known_structure);
        // next one is either restricted or initial structure
        fgets (buffer, sizeof(buffer), oldset_file);    // header "> SSTRAND_ID etc"
        if ( strstr (buffer, "_") != NULL)    // it was the restricted line
        {                        
            strcpy (restricted, buffer);
            fprintf (newset_file, "%s", buffer);
            fgets (initial_structure, sizeof(buffer), oldset_file);    // empty line
            fprintf (newset_file, "%s", initial_structure);
        }        
        else    // it was initial_structure
        {
            restricted[0] = '\0';
            strcpy (initial_structure, buffer);
            fprintf (newset_file, "%s", initial_structure);
        }        
        for (i=0; i < strlen (sequence); i++)
        {
            if(sequence[i] == '\n')
            {
                sequence[i] = '\0';
                known_structure[i] = '\0';
                initial_structure[i] = '\0';
                break;
            }
        }
        for (i=0; i < strlen (restricted); i++)
        {
            if(restricted[i] == '\n')  restricted[i] = '\0';
        }                                     
        
        double known_energy = get_energy (sequence, known_structure);  
        printf ("Known en is %.2lf\n", known_energy);
        double initial_energy = get_energy (sequence, initial_structure);     
        printf ("initi en is %.2lf\n", initial_energy);                 
        
        
        // predict the new structure, write it in the new set file, and measure its accuracy        

        int num_suboptimals = 1;   // Cristina   
        
        //printf ("\nS: %s\nR: %s\n", sequence, restricted);
        if (strcmp (restricted, "") == 0)
        {
            if (UNCONSTRAINED_STRUCTURES == 0)
            {
                printf ("ERROR! constant UNCONSTRAINED_STRUCTURES is 0, but there are UNrestricted structures in your data set!\n");
                exit(1);
            }
            #if (UNCONSTRAINED_STRUCTURES == 1)

                #if (PRINT_SUBOPTIMALS == 0)
                energy = FUNCTION_PREDICTOR (sequence, new_structures[0]);  //, seqname);            
                #elif (PRINT_SUBOPTIMALS == 1)
                energy = FUNCTION_PREDICTOR (sequence, new_structures, energies, num_suboptimals); 
                #endif

            #endif
        }
        else
        {
            if (CONSTRAINED_STRUCTURES == 0)
            {
                printf ("ERROR! constant CONSTRAINED_STRUCTURES is 0, but there are restricted structures in your data set!\n");
                exit(1);
            }
            #if (CONSTRAINED_STRUCTURES == 1)
                #if (PRINT_SUBOPTIMALS == 0)
                energy = FUNCTION_PREDICTOR_CONSTRAINED (sequence, restricted, new_structures[0]);
                #elif (PRINT_SUBOPTIMALS == 1)
                energy = FUNCTION_PREDICTOR_CONSTRAINED (sequence, restricted, new_structures, energies, num_suboptimals);
                #endif

            #endif
        }                

        // copy the rest of the structures into the new set file
        while (!feof(oldset_file))
        {
            fgets (buffer, sizeof(buffer), oldset_file);    // some other structure
            if (buffer[0] == '\n') break;
            buffer [strlen (sequence)] = '\0';
            fprintf (newset_file, "%s\n", buffer);          
        }
            
        fprintf (newset_file, "%s\n", new_structures[0]);
        //add the new line after I write the results
        //fprintf (newset_file, "\n");            
        fflush (newset_file);
       
        sens_initial_train = FUNCTION_COMPUTE_SENSITIVITY (known_structure, initial_structure);
        ppv_initial_train = FUNCTION_COMPUTE_PPV (known_structure, initial_structure);

        // print results for optimal structure
        sens_new_train = FUNCTION_COMPUTE_SENSITIVITY (known_structure, new_structures[0]);
        ppv_new_train = FUNCTION_COMPUTE_PPV (known_structure, new_structures[0]);

        double initial_fmeasure = 0.0;
        double new_fmeasure = 0.0;
        if (sens_initial_train+ppv_initial_train != 0)
            initial_fmeasure = 2*sens_initial_train*ppv_initial_train/(sens_initial_train+ppv_initial_train);
        if (sens_new_train+ppv_new_train != 0)
            new_fmeasure = 2*sens_new_train*ppv_new_train/(sens_new_train+ppv_new_train);
            
        // sanity check
        double new_energy = get_energy (sequence, new_structures[0]);
        if (fabs (energy - new_energy) > 0.01)
        {
            printf ("energy = %.2lf and new_energy = %.2lf should be the SAME!\n", energy, new_energy);
            exit(1);
        }
        fprintf (newset_file, "Energy of known structure: %.2lf\n", known_energy);
        fprintf (newset_file, "Initial sens: %.3lf, Initial ppv: %.3lf, Initial F-measure: %.3lf, Initial Energy: %.2lf\n", sens_initial_train, ppv_initial_train, initial_fmeasure, initial_energy);
        fprintf (newset_file, "New DP  sens: %.3lf, New DP  ppv: %.3lf, New DP  F-measure: %.3lf, New DP  Energy: %.2lf\n", sens_new_train, ppv_new_train, new_fmeasure, energy);

        double new_fmeasure_average_suboptimals = new_fmeasure;
        double new_fmeasure_best_suboptimal = new_fmeasure; // the one for the optimal structure (found above)
        int best_suboptimal = 0;

        if (PRINT_SUBOPTIMALS == 1)
        {
            int counter = 1;
            for (counter = 1; counter < num_suboptimals; counter++)
            {
                if (new_structures[counter] == '\0')
                        break;  // no more suboptimal structures

                fprintf (newset_file, "%s\n", new_structures[counter]);

                sens_new_train = FUNCTION_COMPUTE_SENSITIVITY (known_structure, new_structures[counter]);
                ppv_new_train = FUNCTION_COMPUTE_PPV (known_structure, new_structures[counter]);

                new_fmeasure = 0.0;
                if (sens_new_train+ppv_new_train != 0)
                        new_fmeasure = 2*sens_new_train*ppv_new_train/(sens_new_train+ppv_new_train);
                fprintf (newset_file, "New DP  sens: %.3lf, New DP  ppv: %.3lf, New DP  F-measure: %.3lf, New DP  Energy: %.2lf\n", sens_new_train, ppv_new_train, new_fmeasure, energies[counter]);

                new_fmeasure_average_suboptimals += new_fmeasure;

                if (new_fmeasure_best_suboptimal < new_fmeasure)
                {
                        new_fmeasure_best_suboptimal = new_fmeasure;
                        best_suboptimal = counter;
                }
            }
            new_fmeasure_average_suboptimals /= counter;
        }
        fprintf (newset_file, "New DP  F-measure-avg: %.3lf, New DP  F-measure-best: %.3lf, New DP  best-subopt: %d, New DP  num-subopt: %d\n", new_fmeasure_average_suboptimals, new_fmeasure_best_suboptimal, best_suboptimal, num_suboptimals-1);
        fprintf (newset_file, "\n");    
        
        j++;
        
        fgets (buffer, sizeof(buffer), oldset_file);    // header "> SSTRAND_ID etc"
    }
    fclose (oldset_file);
    fclose (newset_file);

    return 0;  
}


